"""Define the Dataset class

A Dataset is a collection of Item instances, each with a unique identifier,
prompt, response, etc.

>>> from commoneval import dataset, DATAPATH
# create one manually: not the typical case
>>> from datetime import date
>>> ds = dataset.Dataset("a3d", created=date.today(), creator="Biblica", description="A test",
         source="head", subject="test", hasPart=["a3d_000.jsonl", "a3d_001.jsonl"])
>>> ds
<Dataset(a3d, test)>
>>> ds.asdict()
{'identifier': 'a3d', 'contributor': None, 'created': '2025-05-09', 'creator': 'Biblica', 'datePublished': None, 'description': 'A test', 'hasPart': ['a3d_000.jsonl', 'a3d_001.jsonl'], 'language': 'eng', 'license': 'CC-BY-NC-SA-4.0', 'publisher': None, 'source': 'head', 'sourceProcess': '', 'subject': 'test', 'taskPrompt': '', 'title': '', 'version': '1.0'}

# write to file
>>> with (DATAPATH / "a3d.yaml").open("w") as fp:
...   ds.write_yaml(fp)

# read a dataset from a file: this is the typical case
>>> with (DATAPATH / "a3d.yaml").open() as fp:
...   ds2 = dataset.Dataset.read_yaml(fp)
>>> ds2 == ds
True
# read items into a dataset
>>> ds.read_items()


"""

from dataclasses import dataclass, field
from datetime import date
from pathlib import Path
import re
from typing import Any, IO, Optional

import jsonlines
import yaml

from commoneval import DATAPATH
from commoneval.item import Item


@dataclass
class Dataset:
    # by convention, the collection of Items is in a file whose name
    # is derived from the identifier. If multiple files are used, they
    # should be named in hasPart.
    identifier: str
    # when initially created
    created: date
    # who created the dataset. "Gloo" or "Biblica" are high-likelihood values
    creator: Any
    # an account of the content of the dataset
    description: str
    # one or more files that together comprise the dataset
    hasPart: list[str]
    # the source of this data
    source: str
    # perhaps these values should come from a controlled vocabulary someday
    subject: str
    # -- optional attributes
    # individuals who contributed to the dataset
    contributor: Optional[Any] = None
    # when the dataset was published (if published)
    datePublished: Optional[date] = None
    # populate with read_items()
    items: list[Item] = field(default_factory=list)
    # ISO 639-3 code
    language: str = "eng"
    # should be a standard license identifier
    license: str = "CC-BY-NC-SA-4.0"
    # if the license isn't clear or is not a standard one, here's a
    # place to put comments
    licenseNotes: str = ""
    # which version of the metadata spec this dataset and its items conform to
    metadataVersion: str = "3.2"
    # only specify if different from creator
    publisher: Optional[Any] = None
    # how the items were generated
    sourceProcess: str = ""
    taskPrompt: str = ""
    title: str = ""
    version: str = "1.0"
    _localPath: Optional[Path] = None
    _extension: str = "jsonl"
    _slugify_regex: str = r"[^-a-zA-Z0-9_.]+"

    def __post_init__(self) -> None:
        """Check values after initialization."""
        assert re.fullmatch(
            r"[-a-zA-Z0-9_.]+", self.identifier
        ), f"Identifier {self.identifier} has invalid characters."
        # check that the created date is in the past
        assert self.created <= date.today(), "Created date must be in the past."
        # if a single file, identifier+extension; if multiple files, identifier_000.jsonl, etc.
        if len(self.hasPart) == 1:
            assert (
                self.hasPart[0] == f"{self.identifier}.{self._extension}"
            ), f"File name {self.hasPart[0]} does not match single file convention for identifier {self.identifier!r}."
        else:
            for i, part in enumerate(self.hasPart):
                assert (
                    part == f"{self.identifier}_{str(i).zfill(3)}.{self._extension}"
                ), f"File name {part} does not match multiple file convention for identifier {self.identifier!r}."
        if self.datePublished:
            # check that the date published is in the past
            assert (
                self.datePublished <= date.today() and self.datePublished >= self.created
            ), "Date published must be in the past and after created."
        self._localPath = DATAPATH / self.language / self.identifier

    def __repr__(self) -> str:
        """Return a string representation of the Dataset."""
        return f"<Dataset({self.identifier!r}, {self.subject!r})>"

    def __len__(self) -> int:
        """Return the number of items in the Dataset."""
        return len(self.items)

    def asdict(self) -> dict[str, Any]:
        """Return the Dataset as a dictionary for serialization."""
        return {
            "identifier": self.identifier,
            "contributor": self.contributor,
            "created": self.created.isoformat(),
            "creator": self.creator,
            "datePublished": self.datePublished.isoformat() if self.datePublished else None,
            "description": self.description,
            "hasPart": self.hasPart,
            "language": self.language,
            "license": self.license,
            "licenseNotes": self.licenseNotes,
            "metadataVersion": self.metadataVersion,
            "publisher": self.publisher,
            "source": self.source,
            "sourceProcess": self.sourceProcess,
            "subject": self.subject,
            "taskPrompt": self.taskPrompt,
            "title": self.title,
            "version": self.version,
        }

    @staticmethod
    def fromdict(data: dict[str, Any]) -> None:
        """Load the Dataset from a dictionary."""
        safedata = data.copy()
        for key, value in safedata.items():
            if key in ["created", "datePublished"] and value:
                newvalue = date.fromisoformat(value)
                safedata[key] = newvalue
        return Dataset(**safedata)

    @staticmethod
    def read_yaml(fp: IO[str]) -> "Dataset":
        """Read the Dataset metadata from a YAML file.

        Does not read the items.
        """
        ydict = yaml.safe_load(fp)
        return Dataset.fromdict(ydict)

    def write_yaml(self, fp: IO[str]) -> None:
        """Write the Dataset metadata to a YAML file."""
        # write to file
        fp.write(yaml.dump(self.asdict(), sort_keys=False))

    def read_items(self, basepath: Optional[Path] = None) -> None:
        """Read the items from the dataset files."""
        if not basepath:
            basepath = self._localPath
        # this drops any existing items
        self.items = []
        for part in self.hasPart:
            with jsonlines.open(basepath / part, "r") as reader:
                for obj in reader:
                    self.items.append(Item(**obj))
            print(f"Read {len(self.items)} items from {part}")
        return None



"""Infrastructure for scoring recall, precision, and F1 score."""


def precision(true_positives: int, false_positives: int) -> float:
    denom = true_positives + false_positives
    return true_positives / denom if denom else 0.0


def recall(true_positives: int, false_negatives: int) -> float:
    denom = true_positives + false_negatives
    return true_positives / denom if denom else 0.0


def f1(recall: float, precision: float) -> float:
    denom = precision + recall
    return ((2 * precision * recall) / denom) if denom else 0.0



"""Compute GLEU score for two sequences.

>>> from commoneval import gleu
>>> hyp1 = ['It', 'is', 'a', 'guide', 'to', 'action', 'which', 'ensures', 'that', 'the', 'military', 'always', 'obeys', 'the', 'commands', 'of', 'the', 'party']
>>> ref1a = ['It', 'is', 'a', 'guide', 'to', 'action', 'that', 'ensures', 'that', 'the', 'military', 'will', 'forever', 'heed', 'Party', 'commands']
>>> ref1b = ['It', 'is', 'the', 'guiding', 'principle', 'which', 'guarantees', 'the', 'military', 'forces', 'always', 'being', 'under', 'the', 'command', 'of', 'the', 'Party']
>>> gleu(hyp1, ref1a)
{'true_positive': 8, 'false_positive': 7, 'false_negative': 9}
>>> gleu(hyp1, ref1b)
{'true_positive': 3, 'false_positive': 14, 'false_negative': 14}



"""

from collections import Counter


def gleu(seq1: tuple[str], seq2: tuple[str], n: int = 2) -> dict[str, int]:
    def ngrams(seq, n):
        return [tuple(seq[i : i + n]) for i in range(len(seq) - n + 1)]

    ngrams1 = Counter(ngrams(seq1, n))
    ngrams2 = Counter(ngrams(seq2, n))

    true_positive = 0
    false_positive = 0
    false_negative = 0

    for ng in ngrams1:
        if ng in ngrams2:
            true_positive += min(ngrams1[ng], ngrams2[ng])
            if ngrams1[ng] > ngrams2[ng]:
                false_negative += ngrams1[ng] - ngrams2[ng]
            elif ngrams2[ng] > ngrams1[ng]:
                false_positive += ngrams2[ng] - ngrams1[ng]
        else:
            false_negative += ngrams1[ng]

    for ng in ngrams2:
        if ng not in ngrams1:
            false_positive += ngrams2[ng]

    return {
        "true_positive": true_positive,
        "false_positive": false_positive,
        "false_negative": false_negative,
    }



from pathlib import Path

ROOT = Path(__file__).parent.parent
DATAPATH = ROOT / "data"
# where English datasets go
DATAENGPATH = ROOT / "data/eng"
SRCPATH = ROOT / "src"

__all__ = [
    "ROOT",
    "DATAPATH",
    "DATAENGPATH",
    "SRCPATH",
]



"""Define the Item class.

An Item is a single pairing of response and prompt for LLM evaluation, along with other metadata.

See tests/src/test_item.py for examples of how to use this class.

"""

from dataclasses import dataclass, field
import json
import re
import sys
from typing import Any, IO, Type

try:
    from enum import StrEnum  # Available in Python 3.11+
except ImportError:
    # Fallback for Python < 3.11
    from enum import Enum

    class StrEnum(str, Enum):
        pass


# deal with version differencs for StrEnum value checking
def is_valid_enum_value(value: str, enum_class: Type[StrEnum]) -> bool:
    """Check if a value is a valid member of an enum class."""
    if sys.version_info >= (3, 12):
        # Python 3.12+ allows direct containment
        return value in enum_class
    else:
        # Python 3.11 and earlier require manual check
        return value in enum_class._value2member_map_


# enumerate values for moddality
class Modality(StrEnum):
    """Enumerates the types of responses that the prompt expects."""

    # closed-set responses
    #
    BOOLEAN = "boolean"
    CHOICEOF2 = "choiceof2"
    CHOICEOF3 = "choiceof3"
    CHOICEOF4 = "choiceof4"
    CHOICEOF5 = "choiceof5"
    TERNARY = "ternary"
    # open-ended responses
    CLOZE = "cloze"
    SINGLEVALUE = "singlevalue"
    SHORTPROSE = "shortprose"
    LONGPROSE = "longprose"


class Ternary(StrEnum):
    """Enumerate strings representing boolean values plus 'Unknown'."""

    TRUE = "True"
    FALSE = "False"
    UNKNOWN = "Unknown"


@dataclass
class Item:
    """An Item is a single pairing of response and prompt for LLM evaluation, along with other metadata."""

    # The ID of the item
    identifier: str
    # identifies the type of response that the prompt expects
    modality: Modality
    # The prompt used to generate the response
    prompt: str
    # The expected 'gold standard' response: LLM output is compared to this
    response: bool | str
    # optional values below here
    # Content that justifies the reference answer
    support: str = ""
    # Directions for how the LLM should respond to the prompt
    taskPrompt: str = ""
    # The difficulty of the item, from 0.0 to 1.0
    difficulty: float = 0.0
    # stash other data here if needed
    _otherargs: dict[str, Any] = field(default_factory=dict)
    _choiceof2values: set[str] = field(default_factory=lambda: {"A", "B"})
    _choiceof3values: set[str] = field(default_factory=lambda: {"A", "B", "C"})
    _choiceof4values: set[str] = field(default_factory=lambda: {"A", "B", "C", "D"})
    _choiceof5values: set[str] = field(
        default_factory=lambda: {"A", "B", "C", "D", "E"}
    )

    def __post_init__(self):
        """Post-initialization checks for the Item class."""
        assert re.fullmatch(
            r"[-a-zA-Z0-9_.]+", self.identifier
        ), f"Identifier {self.identifier} has invalid characters."
        if self.modality != Modality.BOOLEAN:
            assert len(self.response) > 0, "Response is empty."
        match self.modality:
            # closed-set responses
            case Modality.BOOLEAN:
                assert self.response in {
                    True,
                    False,
                }, "Response is not a valid boolean."

            case Modality.CHOICEOF2:
                assert (
                    self.response in self._choiceof2values
                ), "Response is not a valid choice of 2."
            case Modality.CHOICEOF3:
                assert (
                    self.response in self._choiceof3values
                ), "Response is not a valid choice of 3."
            case Modality.CHOICEOF4:
                assert (
                    self.response in self._choiceof4
                ), "Response is not a valid choice of 4."
            case Modality.CHOICEOF5:
                assert (
                    self.response in self._choiceof5
                ), "Response is not a valid choice of 5."
            case Modality.TERNARY:
                assert is_valid_enum_value(
                    self.response, Ternary
                ), "Response is not a valid ternary value."
            # open-ended responses
            case Modality.CLOZE:
                assert "___" in self.prompt, "Prompt is missing ___ cloze indicator."
            # nothing to check for other modalities
        assert 1.0 >= self.difficulty >= 0.0, "Difficulty must be between 0.0 and 1.0."

    def __repr__(self) -> str:
        """Return a string representation of the Item."""
        if self.modality != Modality.BOOLEAN:
            if len(self.prompt) > 20:
                promptstr = self.prompt[:17] + "..."
            else:
                promptstr = self.prompt
            if len(self.response) > 20:
                responsestr = self.response[:17] + "..."
            else:
                responsestr = self.response
        else:
            promptstr = self.prompt
        responsestr = self.response
        return f"<Item({self.identifier!r}, {self.modality}): {promptstr!r}->{responsestr!r}>"

    def asdict(self) -> dict[str, Any]:
        """Return the Item as a dictionary for serialization."""
        outdict = {
            "identifier": self.identifier,
            "modality": self.modality.value,
            "prompt": self.prompt,
            "response": self.response,
        }
        if self.support:
            outdict["support"] = self.support
        if self.taskPrompt:
            outdict["taskPrompt"] = self.taskPrompt
        return outdict

    def write_jsonline(self, fp: IO[str]) -> str:
        """Return the Item as a JSONL string."""
        json.dump(self.asdict(), fp)
        fp.write("\n")


# Need a subclass for multiple responses: assume for base Item that there's a single response
# need to think through lists or dicts for responses


# Subclass of Item for templated items
# THIS NEEDS WORK: parking code here for now
@dataclass
class TemplatedItem(Item):
    """A subclass of Item for templated items."""

    # if the prompt is templated, the values for instantiating the template
    promptVariables: dict[str, str] = field(default_factory=dict)
    # if the prompt is templated, the values for the responses
    responseVariables: dict[str, str] = field(default_factory=dict)

    def __post_init__(self) -> None:
        """Post-initialization checks for the TemplatedItem class."""
        super().__post_init__()
        # check that the prompt variables are valid
        for key in self.promptVariables:
            assert key in self.get_template_keys(), f"Invalid prompt variable {key}."
        # check that the response variables are valid
        for key in self.responseVariables:
            assert key in self.get_template_keys(), f"Invalid response variable {key}."
        if "{" in self.prompt:
            assert self.isTemplated, "Prompt is templated but isTemplated is False"
            assert len(self.get_template_keys()) == len(
                self.response
            ), "Prompt is templated but number of keys does not match number of responses"
            assert bool(
                self.promptVariables
            ), "Prompt is templated but no prompt variables provided"
            assert bool(
                self.responseVariables
            ), "Prompt is templated but no response variables provided"

    def get_template_keys(self) -> set[str]:
        """If the prompt is templated, return the keys in the template.

        This identifies brackets in the string, ensures they are
        correctly paired and not empty, and returns the set of
        keys. Any duplicate keys will be treated the same.

        If the prompt is not templated, return an empty set.

        """
        if not self.isTemplated:
            return set()
        else:
            keys = set()
            stack = []
            for i, c in enumerate(self.prompt):
                if c == "{":
                    stack.append(i)
                elif c == "}":
                    if not stack:
                        raise ValueError(f"Unmatched '}}' at index {i} in prompt.")
                    start = stack.pop()
                    key = self.prompt[start + 1 : i]
                    if not key:
                        raise ValueError(
                            f"Empty key in prompt at indices {start} to {i}."
                        )
                    keys.add(key)
            if stack:
                raise ValueError(f"Unmatched '{{' at indices {stack} in prompt.")
            return keys

    def generate_prompts(self) -> list[str]:
        """Generate the prompts from the template and the variables."""
        return [
            self.prompt.replace("{" + key + "}", value)
            for index, key in list(enumerate(self.get_template_keys()))
            for value in self.promptVariables[key]
        ]
